.include "nes.h"
.include "ram.h"
.include "float.h"

.segment "BSS"
; Floating-point registers
fpa: .res DFPPREC+1
fpb: .res DFPPREC+1
fpc: .res DFPPREC+1
fpd: .res DFPPREC+1
fpx: .res 4

DFPREG_A = 0
DFPREG_B = <(fpb-fpa)
DFPREG_C = <(fpc-fpa)
DFPREG_D = <(fpd-fpa)

fpaexp = fpa+DFPPREC
fpbexp = fpb+DFPPREC
fpcexp = fpc+DFPPREC
fpdexp = fpd+DFPPREC

.exportzp DFPPREC
floattest_fpa = fpa
.export floattest_fpa

.segment "CODE"

.proc dfpdump_a
  ldx #0
  stx fpx+2
loop:
  stx fpx+3
  lda fpa,x
  jsr bcd8bit
  ora #'0'
  sta fpx+1
  lda $00
  ora #'0'
  sta fpx+0
  lda #>fpx
  ldy #<fpx
  jsr con_puts
  ldx fpx+3
  inx
  cpx #DFPPREC
  bcc loop
  lda fpa+DFPPREC
puthex:
  pha
  lsr a
  lsr a
  lsr a
  lsr a
  jsr hdig
  sta fpx+1
  pla
  and #$0F
  jsr hdig
  sta fpx+2
  lda #0
  sta fpx+3
  lda #'$'
  sta fpx+0
  lda #>fpx
  ldy #<fpx
  jmp con_puts
  
hdig:
  cmp #10
  bcc :+
  adc #'A'-'9'-2
:
  adc #'0'
  rts
.endproc

con_puthex = dfpdump_a::puthex

;;
; Loads register X with the DFP representation of A * 100^Y.
.proc dfpfromu8
  sta fpa+1,x
  tya
  clc
  adc #129
  sta fpa+DFPPREC,x
  lda #0
  .repeat ::DFPPREC-2,I
    sta fpa+2+I,x
  .endrepeat
  sta fpa+0,x
  lda fpa+1,x
loop:
  cmp #100
  bcc done
  inc fpa+0,x
  sbc #100
  bcs loop
done:
  sta fpa+1,x
  rts
.endproc

;;
; Finds the smallest slice of the coefficient that contains all
; nonzero digit pairs.
; Used by normalization, the half-normalization done in addition,
; and metric formatting.
; @param X which register
; @return A=first nonzero, Y=first trailing zero
.proc dfpnzrange
nzmin = 0
nzmax = 1

  ldy #$FF
  sty nzmax
  sty nzmin
  iny  
loop:
  lda fpa+0,x
  beq byteiszero
  sty nzmax
  bit nzmin
  bpl alreadyfoundfirst
  sty nzmin
alreadyfoundfirst:
byteiszero:
  inx
  iny
  cpy #DFPPREC
  bcc loop
  lda nzmin
  ldy nzmax
  iny
  rts
.endproc

;;
; Normalizes a floating point number in register X.
; @return $0002 = pointer to other number
.proc dfpnorm
leadingcount = 0
copyremain = 1
numbase = 2
  stx numbase
  jsr dfpnzrange
  ; A = first nonzero, Y = first trailing zero, Z flag reflects Y
  bne nonzero
  ldx numbase
  lda #0
  sta fpa+DFPPREC,x
already_norm:
  rts
nonzero:
  cmp #0
  beq already_norm
  ldx numbase
  sta leadingcount
  clc
  adc numbase
  tay

  ; Private entry point, assuming the following:
  ; $0000 = left shift count
  ; X = base address of number
  ; Y = $0000 + X
leftshift_entry:
  lda #DFPPREC
  sec
  sbc leadingcount
  sta copyremain
copyloop:
  lda fpa,y
  sta fpa,x
  inx
  iny
  dec copyremain
  bne copyloop
  
  ; Fill remaining bytes in register with 0
  ldy leadingcount
  lda #0
incloop:
  sta fpa,x
  inx
  dey
  bne incloop

  ; And correct the exponent
  sec
  lda fpa,x
  sbc leadingcount
  sta fpa,x
  rts
.endproc

;;
; Copies register D to register selected by X.
.proc dfpmovd
  ldy #<fpd
  lda #>fpd
  bne dfpload
.endproc
;;
; Copies register C to register selected by X.
.proc dfpmovc
  ldy #<fpc
  lda #>fpc
  bne dfpload
.endproc
;;
; Copies register B to register selected by X.
.proc dfpmovb
  ldy #<fpb
  lda #>fpb
  bne dfpload
.endproc
;;
; Copies register A to register selected by X.
.proc dfpmova
  ldy #<fpa
  lda #>fpa
.endproc
;;
; Copies floating-point number at AAYY to register selected by X.
.proc dfpload
  sty 0
  sta 1
.endproc
;;
; Copies floating-point number whose address is in $00-$01
; to register selected by X.
.proc dfpload0
  ldy #0
loop:
  lda (0),y
  sta fpa,x
  inx
  iny
  cpy #DFPPREC+1
  bcc loop
  rts
.endproc

;;
; Copies floating-point register selected by X to address AAYY.
.proc dfpstore
  sty 0
  sta 1
.endproc
;;
; Copies floating-point register selected by X to address in $00-$01.
.proc dfpstore0
  ldy #0
loop:
  lda fpa,x
  sta (0),y
  inx
  iny
  cpy #DFPPREC+1
  bcc loop
  rts
.endproc

.proc dfpformat_a
highdigits = $00
lastnz = $01
dstlo  = $02
dsthi  = $03
wholedp = $04
digs = $05

  sta dsthi
  sty dstlo
  ldx #0
  jsr dfpnzrange
  bne nonzero
  ; zero: that's easy
  ldy #0
  lda #'0'
  sta (dstlo),y
  tya
  iny
  sta (dstlo),y
  rts
nonzero:
  sty lastnz
  ldy #0  ; Y: index into string
  ldx #0  ; X: index into digit pairs
  lda fpaexp
  sec
  sbc #127
  sta wholedp
  bcc nowholedp
  beq nowholedp
wholedp_loop:
  lda fpa,x
  jsr bcd8bit
  pha
  tya
  ora highdigits
  beq wholedp_nohi
  lda highdigits
  ora #'0'
  sta (dstlo),y
  iny
wholedp_nohi:
  pla
  bne wholedp_yeslo
  cpy #0
  beq wholedp_nolo
wholedp_yeslo:
  ora #'0'
  sta (dstlo),y
  iny
wholedp_nolo:
  inx
  cpx wholedp
  bcs wholedp_done
  cpx #6
  bcc wholedp_loop

  ; if more than 12 digits, fill zeroes
  lda #'0'
wholedp_zerotopoint:
  sta (dstlo),y
  iny
  sta (dstlo),y
  iny
  inx
  cpx wholedp
  bcc wholedp_zerotopoint
wholedp_done:
nowholedp:
  cpy #0
  bne nozerobeforepoint
  lda #'0'
  sta (dstlo),y
  iny
nozerobeforepoint:

  ; We're done printing the whole number part.  Now on to decimals.
  cpx lastnz
  bcs terminate
  lda #'.'
  sta (dstlo),y
  iny

  ; Print leading zeroes if smaller than .01
  lda wholedp
  bpl dec_render
  lda #'0'
decloop:
  sta (dstlo),y
  iny
  sta (dstlo),y
  iny
  inc wholedp
  bne decloop
dec_render:
  lda fpa,x
  jsr bcd8bit
  pha
  lda highdigits
  ora #'0'
  sta (dstlo),y
  iny
  pla
  ora #'0'
  sta (dstlo),y
  iny
  inx
  cpx lastnz
  bcc dec_render

  ; Chop trailing 0
  cmp #'0'
  bne terminate
  dey

  ; Terminate the string with a NUL and finish
terminate:
  lda #0
  sta (dstlo),y
  rts
.endproc

;;
; Beautify, using commas.
; @param AY where to write the nul-terminated string
; @param X maximum length of string
; @return X: prefix type (0=none, 1=k, 2=M, ..., 8: E); Y: length
.proc dfpbeautify_a
i      = $00
highdigits = $00
lastnz = $01
dstlo  = $02
dsthi  = $03
wholedigits = $04
tillcomma = $05
pfxtype = $06
maxlen = $07
lastwholedigit = $08  ; if an odd SI prefix splits something
  sta dsthi
  sty dstlo
  stx maxlen
  ldx #$FF
  stx lastwholedigit
  inx
  stx pfxtype
  jsr dfpnzrange
  bne nonzero
  ; zero: that's easy
  ldy #0
  lda #'0'
  sta (dstlo),y
  tya
  iny
  sta (dstlo),y
  tax
  rts
nonzero:
  sta i
  sty lastnz
  tax        ; X = index into decimal pairs
  ldy #0     ; Y = index into output

  lda fpaexp
  sec
  sbc i
  sec
  sbc #127
  bcs positive_digits
  lda #1
  sta wholedigits
  bne nowhole
positive_digits:
  asl a
  sta wholedigits

  ; If lead digit is less than 10, one fewer whole digit
  lda fpa,x
  cmp #10
  bcs :+
  dec wholedigits
:
nowhole:

  ; calculate how many commas: (wholedigits - 1) // 3
  ; this determines what metric prefix to use
  sec
  lda wholedigits
  sbc #1

  .repeat 4, I
    cmp #24 >> I
    bcc :+
    sbc #24 >> I
:
    rol pfxtype
  .endrepeat

  ; figure out how many digits before the decimal point or comma
  ; = wholedigits - num_commas / 3
  lda pfxtype
  asl a
  adc pfxtype
  sec
  eor #$FF
  adc wholedigits
  sta tillcomma

  ; If ((maxlen * 3 + 3) // 4) >= wholedigits then the whole
  ; number part will fit without using metric
  lda maxlen
  asl a
  adc maxlen
  adc #3
  lsr a
  lsr a
  cmp wholedigits
  bcs nometric

  ; With SI prefixes, treat everything after the first
  ; thousands point as decimal points
  lda tillcomma
  sta wholedigits
  dec maxlen
  dec maxlen
  bcc wholedigloop

  ; Without SI prefixes, kill commas
nometric:
  lda #0
  sta pfxtype

  ; The whole number part
wholedigloop:
  lda #0
  cpx #DFPPREC
  bcs :+
  lda fpa,x
:
  jsr bcd8bit
  pha
  lda highdigits
  jsr do_wholedigit
  pla
  jsr do_wholedigit
  inx
  lda wholedigits
  beq wholedone
  bpl wholedigloop
wholedone:

  ; If there's space, write digits after decimal point
  tya
  adc #1
  cmp maxlen
  bcs terminate
  bit lastwholedigit
  bpl stillneedlwd
  cpx lastnz
  bcs terminate
stillneedlwd:
  lda #'.'
  sta (dstlo),y
  iny
  lda lastwholedigit
  bmi :+
  sta (dstlo),y
  iny
:
  jmp decimalentry
  
  ; Write decimal digits until out of digits or space
decimalloop:
  lda fpa,x
  inx
  jsr bcd8bit
  pha
  lda highdigits
  ora #'0'
  sta (dstlo),y
  iny
  pla
  cpy maxlen
  bcs terminate
  cpx lastnz
  bcc yesdecimallo
  cmp #0
  beq terminate
yesdecimallo:
  ora #'0'
  sta (dstlo),y
  iny
decimalentry:
  cpy maxlen
  bcs terminate
  cpx lastnz
  bcc decimalloop

  ; Terminate string with a space, the metric prefix, and a NUL
terminate:
  ldx pfxtype
  beq :+
  lda #' '
  sta (dstlo),y
  iny
  lda si_prefixes-1,x
  sta (dstlo),y
  iny
:
  lda #0
  sta (dstlo),y
  rts

  ; Write a comma if necessary, then a digit if available
do_wholedigit:
  bne yeswholedigit
  cpy #0
  beq nowholedigit
yeswholedigit:
  dec wholedigits
  bmi nocomma
  dec tillcomma
  bpl nocomma
  pha
  lda #','
  sta (dstlo),y
  iny
  lda #2
  sta tillcomma
  pla
nocomma:
  ora #'0'
  bit wholedigits
  bpl nosplitdp
  sta lastwholedigit
  rts
nosplitdp:
  sta (dstlo),y
  iny
nowholedigit:
  rts
.endproc

;;
; Multiplies a floating point register by 2.
; @param X which register
; @param C if true, add 1 to lowest
; @return C = true if a 1 was shifted out
.proc dfpdouble
  php
  txa
  clc
  adc #DFPPREC-1
  tax
  ldy #DFPPREC
  plp
loop:
  lda fpa,x
  rol a
  cmp #100
  bcc :+
  sbc #100
:
  sta fpa,x
  dex
  dey
  bne loop
  rts
.endproc

;;
; Divides a floating point register by 2.
; @param X which register
; @param C if true, add 50 to highest digit pair
; @return C = old lowest digit
.proc dfphalve
  ldy #DFPPREC
loop:
  lda #0
  bcc :+
  lda #99
:
  adc fpa,x
  lsr a
  sta fpa,x
  inx
  dey
  bne loop
  rts
.endproc

;;
; Adds 1 unit in a given place to a floating point register.
; @param X which register
; @return C = shifted out
.proc dfpinc
  clc
  txa
  adc #DFPPREC-1
  tax
  ldy #DFPPREC
start:
  inc fpa,x
  lda fpa,x
  cmp #100
  bcc done
  sbc #100
  sta fpa,x
  dex
  dey
  bne start
done:
  rts
.endproc
; For internal use, there's a second entry point to increment only
; the first 'width' bytes of a DFP register:
;
;   ldx #DFPREG_whatever+width-1
;   ldy #width
;   jsr dfpinc::start
;
; dfpadd uses width=DFPPREC-1 when shifting and rounding after
; an addition that carried out
; dfpmul uses width=i after adding a byte pair

;;
; Adds a reg to A.
.proc dfpadd
shiftamt = $00

tmpb = $0F - DFPPREC

  ; Load the other value into the temporary register
  ldy #0
loadloop:
  lda fpa,x
  sta tmpb,y
  inx
  iny
  cpy #DFPPREC+1
  bcc loadloop

  ; Swap to make sure A has the larger exponent
  lda fpa+DFPPREC
  cmp tmpb+DFPPREC
  bcs noswap
  ldx #DFPPREC
swaploop:
  lda tmpb,x
  ldy fpa,x
  sty tmpb,x
  sta fpa,x
  dex
  bpl swaploop
noswap:

  ; If a = 0 return b
  ldx #DFPREG_A
  jsr dfpnzrange
  bne a_nonzero
  ; 0 + b = b
  ldx #DFPPREC
returnbloop:
  lda tmpb,x
  sta fpa,x
  dex
  bpl returnbloop
  rts

  ; Shift fpa as far left as needed and possible:
  ; min(number of leading zeroes, exponent of a - exponent of b)
a_nonzero:
  ora #0
  beq a_fronted
  sta shiftamt
  lda fpa+DFPPREC
  sec
  sbc tmpb+DFPPREC
  beq no_shift_needed
  cmp shiftamt
  bcs :+
  sta shiftamt
:
  ldy shiftamt
  ldx #0
  jsr dfpnorm::leftshift_entry

  ; If the exponents differ after fronting a, round b
a_fronted:
  lda fpa+DFPPREC
  sec
  sbc tmpb+DFPPREC
  beq no_shift_needed
  
  ; If A's exponent is so much bigger than that of B that addition
  ; would not change A, leave A unchanged
  cmp #DFPPREC+1
  bcc :+
  rts
:
  ; A is shift amount, 1 <= A <= DFPPREC
  ; Now point Y at the byte about to be shifted out
  sta shiftamt
  eor #$FF
  adc #DFPPREC+1
  tay
  ; Ready to add tmpb starting at digitpair Y-1 to fpa starting at
  ; position X.  This means Y points at the byte about to be shifted
  ; out, so set the carry if the result would need rounded up.
  lda tmpb,y
  cmp #50
  dey
  bmi skip_adding
  bpl ready_to_add
no_shift_needed:
  clc
  ldy #DFPPREC-1
ready_to_add:
  ldx #DFPPREC-1

  ; Now add the carry and the remaining bytes in tmpb to fpa.
addloop:
  lda fpa,x
  adc tmpb,y
  cmp #100
  bcc :+
  sbc #100
:
  sta fpa,x
  dex
  dey
  bpl addloop
  bcc no_carry_after_add

  ; If there was a carry out after adding, there are two
  ; possibilities.  First, if b's exponent was less, try incrementing
  ; the bytes that haven't been added to yet.
  cpx #0
  bmi skip_adding
  txa
  tay
  iny
  jsr dfpinc::start
  
  ; Finally, if there was a carry out all the way to the upper place
  ; of fpa, we need to round fpa, shift it, and copy it out.
skip_adding:
  bcc no_carry_after_add
  lda fpa+DFPPREC-1
  cmp #50
  ldx #DFPPREC-2
carryoutloop:
  lda fpa,x
  sta fpa+1,x
  dex
  bpl carryoutloop
  lda #1
  sta fpa
  bcc no_carry_after_add
  jmp dfpinc
no_carry_after_add:
  rts
.endproc


;;
; dfpsub: Set A = abs(A - register X) and return the sign.
; dfpcmp: Only return the sign.
; Side effects: both A and the other register are normalized
.proc dfpcmp
  sec
  .byte $90  ; BCC opcode to skip first byte of dfpsub
.endproc
.proc dfpsub
othernum = $06
signret = $07
  clc
  ror signret
  stx othernum
  jsr dfpnorm
  ldx #0
  jsr dfpnorm
  ldx othernum
  ; Compare exponents of normalized numbers
  lda fpa+DFPPREC
  cmp fpa+DFPPREC,x
  bcc signneg
  bne signpos

  ; Compare digit-wise
  ldy #0
cmpdigitwise:
  lda fpa,y
  cmp fpa,x
  bcc signneg
  bne signpos
  inx
  iny
  cpy #DFPPREC
  bcc cmpdigitwise

  ; Numbers are equal
  lda #0
  bit signret
  bmi return0sign
  ldx #DFPPREC
return0loop:
  sta fpa,x
  dex
  bpl return0loop
return0sign:
  lda #0
  rts

signneg:
  lda #$FF
  bmi :+
signpos:
  lda #1
:
  bit signret
  bpl notcmp
  ora #0  ; set NZ according to flag
  rts
notcmp:
  sta signret
  
tmpb = $0F - DFPPREC

  ; Load the nines' complement of the subtrahend (the smaller value)
  ; into a temporary register
  ldy #0
  ldx #0
  clc
  bit signret  ; bit 7 set if fpa is smaller
  bmi mvsmaller
  ldx othernum
mvsmaller:
  lda #99+1
  sbc fpa,x
  sta tmpb,y
  inx
  iny
  cpy #DFPPREC
  bcc mvsmaller
  lda fpa,x
  sta tmpb+DFPPREC

  ; Move the minuend (the larger value) to fpa
  bit signret
  bpl nomvgreater
  ldx othernum
  ldy #0
mvgreater:
  lda fpa,x
  sta fpa,y
  inx
  iny
  cpy #DFPPREC+1
  bcc mvgreater
nomvgreater:

  ; Round if necessary
  ;sec  ; carry is already set from the CMP in the above loops
  ldy #DFPPREC
  lda tmpb+DFPPREC
  sbc fpa+DFPPREC
  beq noround
  ;clc  ; carry is clear because tmpb < fpa
  adc #DFPPREC
  bcs not_noeffect  ; subtrahend too small to have any effect
  rts
not_noeffect:
  tay
  lda tmpb,y
  sta tmpb+DFPPREC
  cmp #50
  lda #0
  ; inbound carry = -1
noround:
  ldx #DFPPREC-1
  dey
  bmi subdone
  
  ; do the subtraction
subloop:
  lda fpa,x
  adc tmpb,y
  cmp #100
  bcc :+
  sbc #100
:
  sta fpa,x
  dex
  dey
  bpl subloop
subdone:
  bcs subdone1

carryoutloop:
  ; At this point we need to subtract 1 from fpa[x] through fpa[0]
  dec fpa,x
  bpl subdone1
  lda #99
  sta fpa,x
  dex
  bne carryoutloop
  dec fpa
  bne subdone1

  ; We have a subtraction that needs to use a guard digit pair.
  ; The guard digit is tmpb+DFPPREC
  ; if it's at least 50 we may need to undo a rounding
  ; TO DO: Let's punt this until it's needed though

subdone1:
  ; Return the sign in A
  lda signret
  rts
.endproc

.proc dfpmul
qsqlo_plus  = $08
qsqlo_minus = $0A
qsqhi_plus  = $0C
qsqhi_minus = $0E
b_base      = fpx+0
bindex      = fpx+1
slo         = fpx+2
shi         = fpx+3
ccoef       = $00
  lda #>qsqlo
  sta qsqlo_plus+1
  sta qsqlo_minus+1
  lda #>qsqhi
  sta qsqhi_plus+1
  sta qsqhi_minus+1
  stx b_base

  ; normalizing first preserves precision in result
  jsr dfpnorm
  ldx b_base
  lda fpa+DFPPREC,x
  bne b_nonzero
return_zero:
  ldy #0
  tya
  tax
  jmp dfpfromu8
b_nonzero:
  ldx #DFPREG_A
  jsr dfpnorm
  lda fpa+DFPPREC
  beq return_zero

  ; Calculate the exponent
  ldx b_base
  lda fpa+DFPPREC,x
  clc
  adc fpa+DFPPREC
  rol ccoef+0  ; ccoef[0]:A = aexp+bexp
  sec
  sbc #127
  sta fpa+DFPPREC
  lda #$01
  and ccoef+0
  sbc #0
  bcc return_zero
  beq not_infinity
  ; FIXME: do something to handle multiply overflow
not_infinity:
  lda fpa+DFPPREC
  beq return_zero  
  ldx #DFPPREC-1
  stx bindex
  lda #0
clearcloop:
  sta ccoef,x
  dex
  bpl clearcloop

  ; Construct each digit of the product
bdploop:

  ; Shift digits down one unit
  ldx #DFPPREC-1
cshiftloop:
  lda ccoef,x
  sta ccoef+1,x
  dex
  bpl cshiftloop
  inx
  stx ccoef
  clc
  lda b_base
  adc bindex
  tax
  lda fpa,x
  beq bdp_is_zero

  ; Build addresses of the quarter square table offsets for this digit pair
  tax
  clc
  adc #<(99+qsqlo)
  sta qsqlo_plus
  txa
  clc
  adc #<(99+qsqhi)
  sta qsqhi_plus
  txa
  eor #$FF
  tax
  sec
  adc #<(99+qsqlo)
  sta qsqlo_minus
  txa
  sec
  adc #<(99+qsqhi)
  sta qsqhi_minus

  ; Now look up the quarter squares for this in all 
  ldx #DFPPREC-1
  clc
adploop:
  bcc :+
  inc ccoef,x
:
  ldy fpa,x
  sec
  lda (qsqlo_plus),y
  sbc (qsqlo_minus),y
  bcs :+
  adc #100
  clc
:
  sta slo
  lda (qsqhi_plus),y
  sbc (qsqhi_minus),y
  sta shi

  ; at this point, shi:slo = a[i] * b[j]
  clc
  lda ccoef+1,x
  adc slo
  cmp #100
  bcc :+
  sbc #100
:
  sta ccoef+1,x
  lda ccoef,x
  adc shi
  cmp #100
  bcc :+
  sbc #100
:
  sta ccoef,x
  dex
  bpl adploop

bdp_is_zero:
  dec bindex
  bpl bdploop
  ; If the result has no leading zero, round it
  ldx #DFPPREC-1
  lda fpa+DFPPREC
  cmp #2
  bcc nlzloop
  lda ccoef+0
  beq lzloop
nlzloop:
  lda ccoef,x
  sta fpa,x
  dex
  bpl nlzloop
  lda ccoef+DFPPREC
  cmp #50
  bcc done
  inx
  jmp dfpinc

  ; So the result has a leading zero.  Use the guard digits.
lzloop:
  lda ccoef+1,x
  sta fpa,x
  dex
  bpl lzloop
  dec fpa+DFPPREC
done:
  rts
.endproc

.proc dfpdiv
  rts
.endproc

.segment "RODATA"
.align 256

; Quarter square tables for multiplication
qsqlo:
  .repeat 298, I
    .byte ((I - 99)*(I - 99)/4) .mod 100
  .endrepeat
; so long as the first 200 entries of the qsqhi table are
; within the same page, it's fine
qsqhi:
  .repeat 298, I
    .byte ((I - 99)*(I - 99)/400)
  .endrepeat

si_prefixes: .byte "kMGTPEZY"


