;
; Title screen and main clickable screen for idle game
; Copr. 2014 Damian Yerrick
;
.include "nes.h"
.include "ram.h"
.include "mbyt.h"
.include "float.h"

.segment "BSS"
transition_timer: .res 1

; Use MMO wordmark until buildings get implemented.
; http://youtu.be/cD69PAIqiYo
USE_MMO_WORDMARK = 1
; If the game has buildings but not upgrades, use the Classic logo.
; http://orteil.dashnet.org/experiments/cookie/
USE_CLASSIC_WORDMARK = 0

.segment "BANK04"
.export title_screen_04, setup_big_cookie_04
.export do_big_cookie_04

;;
; Sets up video memory for the main clickable (the "big cookie").
.proc setup_big_cookie_04
  jsr load_big_cookie_screen
  jmp bankrts
.endproc

;;
; Loads the striped background in nametable A.
.proc load_stripes
  sta PPUADDR
  lda #$80
  sta PPUADDR
  lda #$04
  ldy #11
striperowloop:
  ldx #64
stripetileloop:
  sta PPUDATA
  eor #$04^$01
  dex
  bne stripetileloop
  dey
  bne striperowloop
  rts
.endproc

.proc load_big_cookie_screen
  ldx #0
  stx last_PPUCTRL
  stx PPUMASK
  lda #VBLANK_NMI
  sta PPUCTRL
  lda #$3F
  bit PPUSTATUS
  sta PPUADDR
  stx PPUADDR
loadpalloop:
  lda bigcookie_pal,x
  sta PPUDATA
  inx
  cpx #32
  bcc loadpalloop

  lda #$00
  tay
  ldx #$20
  jsr ppu_clear_nt
  lda #$20
  jsr load_stripes
  lda #titlechr_specs_end-5-titlechr_specs
chrloop:
  pha
  tax
  lda titlechr_specs+4,x  ; length
  sta $08
  lda titlechr_specs+3,x  ; destination address
  sta PPUADDR
  lda titlechr_specs+2,x
  sta PPUADDR
  lda titlechr_specs+1,x  ; source address
  ldy titlechr_specs+0,x
  ldx $08
  jsr load_pb53_blk
  pla
  sec
  sbc #5
  bcs chrloop

  ; Draw big cookie attributes
  lda #$23
  sta PPUADDR
  ldy #$C8
  sty PPUADDR
  ldx #0
attrloop:
  lda bigcookie_attr,x
  sta PPUDATA
  inx
  cpx #48
  bcc attrloop

  ldy #rgn_bigcookie-rectrgnspecs
  jsr draw_rectrgnspec
  ldy #rgn_topbar-rectrgnspecs
  jsr draw_rectrgnspec
  ldy #rgn_ticker-rectrgnspecs
  jsr draw_rectrgnspec

  rts
.endproc

.proc load_title_extras
  lda #$00
  sta last_PPUCTRL
  sta PPUMASK
  tay
  ldx #$24
  jsr ppu_clear_nt
  lda #$24
  jsr load_stripes

  lda #$08
  sta PPUADDR
  lda #$00
  sta PPUADDR
  lda #>wordmark_tiles_pb53
  ldy #<wordmark_tiles_pb53
  ldx #128*40/512
  jsr load_pb53_blk

  ; Clear area behind status window
  lda #$0D
  ldx #$00
  sta PPUADDR
  stx PPUADDR
  txa
clrtxtloop:
  sta PPUDATA
  sta PPUDATA
  sta PPUDATA
  inx
  bne clrtxtloop

  jsr write_copr_notice
  lda #VBLANK_NMI
  sta PPUCTRL

; Draw "Cookie Clicker" wordmark
dstlo = 0
dsthi = 1
  ldx #$24
  stx dsthi
  ldx #$E2
  stx dstlo
  ldx #0
wordmark_rowloop:
  lda dsthi
  sta PPUADDR
  lda dstlo
  sta PPUADDR
  clc
  adc #32
  bcc :+
  inc dsthi
:
  sta dstlo
  ldy #28
:
  lda wordmark_map,x
  sta PPUDATA
  inx
  dey
  bne :-
  cpx #4*28
  bcc wordmark_rowloop

  ldy #rgn_title_bigcookie-rectrgnspecs
  jsr draw_rectrgnspec
  ldy #rgn_title_topbar-rectrgnspecs
  jsr draw_rectrgnspec
  ldy #rgn_title_ticker-rectrgnspecs
  jsr draw_rectrgnspec

  lda #$27
  sta PPUADDR
  ldy #$C8
  sty PPUADDR
  ldx #0
attrloop:
  lda title_attr,x
  sta PPUDATA
  inx
  cpx #48
  bcc attrloop

  ; to do: load bottom row of perfectCookie into pattern table 1
  ; to display over milk
  
  rts
.endproc

.proc title_screen_04
  jsr load_title_extras
  lda #128
  sta transition_timer
  ; Put the cursor where it goes
  lda #132
  sta cursor_x
  sta cursor_x_target
  lda #188
  sta cursor_y_target
  lda #255
  sta cursor_y

bigcookie:
  jsr load_big_cookie_screen

  ; Load initial cookie count to nametable 1 if title screen
  ; or nametable 0 otherwise
  dfplda cookies
  lda #0
  jsr build_topbarbuf
  bit transition_timer
  bpl :+
  lda #$24
  sta bgup_args+1
:

  ; Enable the new NMI handler 
  lda #BG_ON
  sta last_PPUMASK
  lda #VBLANK_NMI|BG_0000|OBJ_1000|1
  sta last_PPUCTRL

  ; transition_timer is 128 for title screen, 32-2 for transition
  ; from title screen to big cookie, or 0 for big cookie
mainloop:
  lda #VBLANK_NMI|OBJ_1000|BG_0000|0
  bit transition_timer
  bpl :+
  ora #1
:
  sta last_PPUCTRL
  
  ; (Transition only) Set up sprite 0
  lda transition_timer
  bmi have_oam_start
  beq have_oam_start
  clc
  adc #30
  sta OAM+0
  lda #1
  sta OAM+1
  lda #%00100000  ; behind bg
  sta OAM+2
  lda #248
  sta OAM+3
  lda #4
have_oam_start:
  sta oam_used
  jsr draw_main_cursor
  ; to do: update milk
  ; to do: update cursor strip
  ; to do: draw golden cookie
  ; to do: draw deer

  ; Draw stripes behind big cookie
  lda transition_timer
  bpl :+
  lda #32
:
  clc
  adc #80-1  ; top edge of big cookie surround
  tay
  jsr draw_big_cookie_surround

  ldx oam_used
  jsr ppu_clear_oam
  lda #1
  sta oam_ready
  lda #BG_ON|OBJ_ON
  sta last_PPUMASK

  lda nmis
:
  cmp nmis
  beq :-
  jsr read_pads
  jsr move_main_cursor

  ; If in transition, wait for sprite 0 hit
  lda transition_timer
  bmi no_s0wait
  beq no_s0wait
  bit PPUSTATUS
  bvs :-
  lda #$40
:
  bit PPUSTATUS
  beq :-
  lda #$00
  sta PPUADDR
  lda #$80
  sta PPUADDR

  ; Now wait for below the big cookie and move up
  ldx #13
:
  dex
  bne :-
  lda #159
  sec
  sbc transition_timer
  tax
  jsr waitxlines
  dec transition_timer
  dec transition_timer
  lda #$03
  sta PPUADDR
  lda #$00
  sta PPUADDR
  beq notStart
no_s0wait:

  ; Look for A button presses
  lda new_keys
  and #KEY_A
  beq notStart
  lda #112
  bit transition_timer
  bpl :+
  lda #144
:
  jsr big_cookie_hitdetect
  bcs notStart

  ; Start has been pressed.  Switch to main nametable.
  dfpldb cookiesPerClick
  dfplda cookies
  ldx #DFPREG_B
  jsr dfpadd
  dfpsta cookies

  lda #0
  jsr build_topbarbuf
  ldx transition_timer
  bpl notStart
  ldx #32
  stx transition_timer
notStart:
  jmp mainloop
bigcookiedone:
  rtl
.endproc

do_big_cookie_04 = title_screen_04::bigcookie

.proc waitxlines
  lda #86
loop:
  ldy #20
:
  dey
  bne :-
  bit $00
  adc #170
  bcs :+
:
  dex
  bne loop
  rts
.endproc
.assert >*=>waitxlines, error, "waitxlines crosses page boundary"

titlechr_specs:
  .addr topbar_tiles_pb53, $0000
  .byte 128*24/512
  .addr big_cookie_pb53, $0400
  .byte 64*64/512
  .addr big_cookie_surround_pb53, $1100
  .byte 64*16/512
  .addr cursor_sprite_pb53, $1000
  .byte 64*16/512
titlechr_specs_end:

rectrgnspecs:
rgn_title_bigcookie:
  .addr $25CC
  .byte $40, 8, 8, 8
rgn_title_topbar:
  .addr $2442
  .byte $14, 16, 11, 2
rgn_title_ticker:
  .addr $2744
  .byte $D0, 24, 24, 2

rgn_bigcookie:
  .addr $214C
  .byte $40, 8, 8, 8
rgn_topbar:
  .addr $2042
  .byte $14, 16, 11, 2
rgn_ticker:
  .addr $2344
  .byte $D0, 24, 24, 2

.proc draw_rectrgnspec
  ldx #0
copyloop:
  lda rectrgnspecs,y
  sta $02,x
  iny
  inx
  cpx #6
  bcc copyloop
  jmp draw_recttileregion
.endproc
  
.proc draw_big_cookie_surround
ycoord = $00
tilenum = $01
  sty ycoord
  ldx oam_used
  ldy #15
loop:
  lda surround_map,y
  and #$38
  clc
  adc ycoord
  sta OAM,x
  inx
  tya
  clc
  adc #$10
  sta OAM,x
  inx
  lda #0
  sta OAM,x
  inx
  lda surround_map,y
  and #$07
  asl a
  asl a
  asl a
  asl a
  adc #96
  sta OAM,x
  inx
  dey
  bpl loop
  stx oam_used
  rts
surround_map:
  mbyt "00010203 080B 10 18 20 28 3033 38393A3B"
.endproc

;;
; Set carry iff the click was within the big cookie.
.proc big_cookie_hitdetect
  ; was the click within the cookie?
  sec
  sbc cursor_y
  bcs :+
  eor #$FF
:
  cmp #32
  bcs have_c
  sta 0
  lda cursor_x
  sec
  sbc #128
  bcs :+
  eor #$FF
:
  cmp #32
  bcs have_c
  sta 1
  cmp 0
  bcs :+
  lda 0
:
  clc
  adc 0
  adc 1
  cmp #68
have_c:
  rts
.endproc

bigcookie_pal:
  mbyt "0F172737 0F001020 0F122220 0F172712"
  mbyt "0F0F2220 0F161616 0F161616 0F161616"
title_attr:
  mbyt "AAAAAAAAAAAAAAAA"
  mbyt "AAAAAAAAAAAAAAAA"
  mbyt "AAAAAAFAFAAAAAAA"
  mbyt "AAAAAA33CCAAAAAA"
  mbyt "AAAAAAAFAFAAAAAA"
  mbyt "5A5A5A5A5A5A5A5A"
bigcookie_attr:
  mbyt "AAAAAAAAAAAAAAAA"
  mbyt "AAAAAAFAFAAAAAAA"
  mbyt "AAAAAA33CCAAAAAA"
  mbyt "AAAAAAAFAFAAAAAA"
  mbyt "AAAAAAAAAAAAAAAA"
  mbyt "5A5A5A5A5A5A5A5A"
topbar_tiles_pb53:
  .incbin "../obj/nes/topbar_tiles.chr.pb53"
big_cookie_pb53:
  .incbin "../obj/nes/big_cookie.chr.pb53"
big_cookie_surround_pb53:
  .incbin "../obj/nes/big_cookie_surround.chr.pb53"
cursor_sprite_pb53:
  .incbin "../obj/nes/cursor_sprite16.chr.pb53"

.if USE_MMO_WORDMARK
wordmark_tiles_pb53:
  .incbin "../obj/nes/mmo_wordmark_tiles.chr.pb53"
  .res 16, $84
wordmark_map:
  mbyt "0401040104018081828304018081 8283040186878889040104010401"
  mbyt "0401040104019091929304019091 9293040196979899040104010401"
  mbyt "040104010401A0A1A2A3A4A5A0A1 A2A3A4A5A6A7A8A9A4A504010401"
  mbyt "040104010401B0B1B2B3B4B5B0B1 B2B3B4B5B6B7B8B9B4B504010401"
.elseif USE_CLASSIC_WORDMARK
; <orteil>
wordmark_tiles_pb53:
  .incbin "../obj/nes/classic_wordmark_tiles.chr.pb53"
  .res 32, $84
wordmark_map:
  mbyt "8283040104018A01868704010401 82838485868704018A0104010401"
  mbyt "9293909190919A9B96979C9D0401 92939495969798999A9B9C9D9E9F"
  mbyt "A2A3A0A1A0A1AAABA6A7ACAD0401 A2A3A4A5A6A7A8A9AAABACADAEAF"
  mbyt "0401040104010401040104010401 0401040104010401040104010401"
; </orteil>
.else
; <orteil>
wordmark_tiles_pb53:
  .incbin "../obj/nes/wordmark_tiles.chr.pb53"
wordmark_map:
  mbyt "8B8C040104010485868788010401 8B8C8D8E8F010485860104010401"
  mbyt "9B9C909192939495969798999A01 9B9C9D9E9F81829596C0C1C2C3C4"
  mbyt "ABACA0A1A2A3A4A5A6A7A8A9AA01 ABACADAEAF8384A5A6C5C6C7C8C9"
  mbyt "BBBCB0B1B2B3B4B5B6B7B8B9BA01 BBBCBDBEBF898AB5B6CACBCCCDCE"
; </orteil>
.endif

; fixed bank stuff
.segment "CODE"
;;
; Draws a contiguous set of tiles to the 
.proc draw_recttileregion
dst = $02
tileno = $04
pitch = $05
width = $06
height = $07
rowloop:
  lda dst+1
  sta PPUADDR
  lda dst
  sta PPUADDR
  clc
  adc #32
  sta dst
  bcc :+
  inc dst+1
:
  ldx tileno
  ldy width
tileloop:
  stx PPUDATA
  inx
  dey
  bne tileloop
  lda tileno
  clc
  adc pitch
  sta tileno
  dec height
  bne rowloop
  rts
.endproc

